#ifndef __NLOHIST_HH__
#define __NLOHIST_HH__

//STARTHEADER
// $Id: NLOHist.hh 804 2011-10-07 19:10:19Z salam $
//
// Copyright (c) 2007-2011  Gavin Salam, Sebastian Sapeta and Gregory Soyez
//
//----------------------------------------------------------------------
//
//  NLOHist is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  SimpleHist is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with FastJet; if not, write to the Free Software
//  Foundation, Inc.:
//      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------
//ENDHEADER

#include "SimpleHist.hh"
#include<vector>
#include<string>
#include<ostream>
#include<iomanip>

/// Class for carrying out histogramming in NLO calculations
///
/// 
class NLOHist {
public:
  NLOHist() {
    declare("dummy",0.0, 1.0, 10);
  }

  /// initialise a histogram for NLO studies
  NLOHist(const std::string & name, double minv, double maxv, int n) {
    declare(name, minv, maxv, n);
  }

  /// a way to get the histogram produced outside of a constructor
  void declare(const std::string & name, double minv, double maxv, int n) {
    _name = name;
    // initialise _main, _sqr, _tmp
    _main.declare(minv, maxv, n);
    _sqr .declare(minv, maxv, n);
    _tmp .declare(minv, maxv, n);
    // our normalisation
    _nev = 0.0;
    _nev_bad = 0.0;
    _event_is_bad = false;
  }

  /// return the number of events
  double nev() const {return _nev;}
  double nev_bad() const {return _nev_bad;}

  /// add an entry to the histogram
  void add_entry(double v, double weight) {
    // test for NaN
    if (weight != weight) {
      _event_is_bad = true;
      return;
    }
    // otherwise, bin it
    int ibin = _tmp.bin(v);
    if (_tmp[ibin] == 0) _in_use.push_back(ibin);
    _tmp[ibin] += weight;
  }


  /// collect the information binned within the event
  /// and transfer it to the definitive store (_main, _sqr)
  void collate_event() {
    _nev += 1;
    for (unsigned i = 0; i < _in_use.size(); i++) {
      int ibin = _in_use[i];
      // only transfer info from whole event, if there
      // were no bad events in it at all
      if (!_event_is_bad) {
        _main[ibin] += _tmp[ibin];
        _sqr [ibin] += _tmp[ibin]*_tmp[ibin];
      }
      _tmp[ibin]   = 0.0;
    }
    _in_use.clear();

    // if things were bad, reset histograms
    if (_event_is_bad) {
      _event_is_bad = false;
      _nev_bad++;
    }

  }


  /// output the histogram with errors, in a format similar to
  /// Zoltan's
  ///
  /// NB: we here normalise to nev(), but suspect that Zoltan
  ///     instead normalises to nev()-nev_bad(); 
  ///
  ///     we tend to believe that our choice is more correct [though
  ///     the effect should be sub per-mille level, at worst, for
  ///     dipole phase-space]: the normalisation of the "good" bins,
  ///     should not be affected by the fact that in "difficult" bins
  ///     the finite-precision numerics gave a NaN rather than the
  ///     correct answer.
  ///
  void write(std::ostream & ostr, int index) {
    // make sure we're not in the middle of an event
    assert(_in_use.size() == 0);
    ostr << "# " << _name << " (index = " << index << ")"<< std::endl;
    // calculate the errors
    for (unsigned ibin = 0; ibin < _tmp.outflow_size(); ibin++) {
      // this will be used for the error, and contains a factor
      // nev**2 inside the sqrt currently (will be removed when writing)
      _tmp[ibin] = sqrt(std::abs(_sqr[ibin] - _main[ibin]*_main[ibin]/_nev));
    }

    // finally output the histogram with errors
    ostr << std::setprecision(8); // get decent precision
    output(_main, _tmp, &ostr, 1.0/_nev/_tmp.binsize());

    // reset tmp
    _tmp *= 0.0;
  }

  /// return the name given to the histogram
  const std::string & name() const {return _name;}
  const double min() const {return _main.min();}
  const double max() const {return _main.max();}

private:
  
  double      _nev, _nev_bad;
  SimpleHist  _main, _sqr, _tmp;
  std::vector<int> _in_use;
  std::string _name;
  bool        _event_is_bad;
};
#endif // __NLOHIST_HH__

