#!/usr/bin/env perl
#
# Small script that starts with a specific configuration, e.g. Q[1],
# and allows one to determine the configurations (and
# splitting-functions weights) associated with all possible splittings
# of that configuration. 
#
# It works recursively, so that one can consider one splitting, two
# splittings, etc.
#
# The results can be output to mathematica (and this is arguably
# extensible to C++ too).

use warnings;
use strict;

#
my $cmdline = "$0 ".join(" ",@ARGV);
print "# file generated automatically by $cmdline\n";

# set up the configuration - for now the nomenclature is a little
# weird, with a |Q and |G - this is just to help 
# my $config = "|Q[1]";
# my $name = "quark";
my $config = "|G[1]";
my $name = "gluon";

print "# starting configuration: $config \n";

my $config1 = &splitv2($config,"z");
print "# first order $name: \n";
print &cleanupCpp($config1),"\n";

my $config2 = &splitv2($config1,"zp");
# print "# second order $name: \n";
# print &cleanupCpp($config2),"\n";

my $config3 = &splitv2($config2,"zpp");
# print "# third order $name: \n";
# print &cleanupCpp($config3),"\n";

my $config4 = &splitv2($config3,"zppp");
# print "# forth order $name: \n";
# print &cleanupCpp($config4),"\n";

my $config5 = &splitv2($config4,"zpppp");
# print "# forth order $name: \n";
# print &cleanupCpp($config5),"\n";

#----------------------------------------------------------------------
# generate output cleaned up for mathematica
sub cleanupMathematica($) {
    (my $string) = @_;
    my $result = $string." "; # force a copy
    $result =~ s/_([a-z0-9]+)/\[$1\]/g; # replace pij_z -> pij[z], etc.
    $result =~ s/\|//g;                 # eliminate "|", needed only internally
#    $result =~ s/\n//sg;                # eliminate newlines
    $result =~ s/om([a-z]+)/\(1-$1\)/g; # replace omz -> (1-z)
    return $result;
}
#----------------------------------------------------------------------
# generate output cleaned up for c++
sub cleanupCpp($) {
    (my $string) = @_;
    my $result = $string." "; # force a copy
    # $result =~ s/_([a-z0-9]+)/\[$1\]/g; # replace pij_z -> pij[z], etc.
    $result =~ s/\|//g;                 # eliminate "|", needed only internally
    $result =~ s/\+//g;                 # eliminate "+", needed only internally
#    $result =~ s/\n//sg;                # eliminate newlines
    $result =~ s/om([a-z]+)/\(1-$1\)/g; # replace omz -> (1-z)
    return $result;
}


#----------------------------------------------------------------------
# Given a string that involves a series of terms separate by "+", each
# containing sequences like
#
#     Q[x] G[y] Q[z]
#
# replace those sequences by 
#
#    (Max[x,y,z] - 1)
sub setMax($) {
    my ($configs) = @_;
    my @configs = split(/\+/, $configs);
    my $result;

    foreach my $config (@configs) {
        if ($config =~ /^ *$/) {next;}
        my @pieces = split(/\|?[QG]/,$config);
        my $newconfig  = shift @pieces;
        $newconfig .= "(-1+Max[";
        foreach my $piece (@pieces) {
            $piece =~ s/[\[\]]//g;
        }
        $newconfig .= join(", ",@pieces)."])";
        $result .= "+".$newconfig;
    }
    return $result;
}

#----------------------------------------------------------------------
# take a string and for each element perform replacements such as
# |Q[x] -> t * pqq_z*(|Q[x*z] |G[x*omz] - |Q[x])
# writing the result as an expanded sum of terms
sub splitv2($$) {
   my ($configs, $splittingVar) = @_;
   $configs =~ s/\n//sg;
   
   my $result = "";

   my @configs = split(/\+/, $configs);
   foreach my $config (@configs) {
       
     # loop over things starting with |
     my $start = 0;
     while (1) {
         $start = index($config, "|", $start);
         if ($start < 0) {last;}
         my $end   = index($config, "]", $start);
         if ($end < 0) {die "Error, could not find ]";}
  
         my $parton = substr($config, $start, $end-$start+1);
         # get a configuration with that parton stripped out (so we can easily replace it)
         my $strippedconfig = $config;
         substr($strippedconfig, $start, $end-$start+1) = "";
         # ugly code to get "x"
         (my $x = $parton) =~ s/.*\[//;
         $x =~ s/\]//;
         if   ($parton =~ /\|Q/) {
             # the real term
             $result = $result . "+pqq_$splittingVar $strippedconfig |Q[$x*$splittingVar] |G[$x*om$splittingVar]\n";
             # the virtual term - just leaves the original configuration
             $result = $result . "+pqq_$splittingVar $config\n";
         } elsif ($parton =~  /\|G/) {
             # the real Pgg & Pqg terms
             $result = $result . "+pgg_$splittingVar $strippedconfig |G[$x*$splittingVar] |G[$x*om$splittingVar]\n";
             $result = $result . "+pqg_$splittingVar $strippedconfig |Q[$x*$splittingVar] |Q[$x*om$splittingVar]\n";
             # the virtual Pgg & Pqg terms - just leave the original configuration
             $result = $result . "+pgg_$splittingVar $config\n";
             $result = $result . "+pqg_$splittingVar $config\n";
         } else {
             die "Could not process $parton";
         }
         $start = $end;
     }
   }

   # clean up the result to remove redundant multipliers
   $result =~ s/1\*//g;
   return $result;
}
