/// \file example2.cc Example that prints out a wider range of variables than example.cc
///
/// Usage:
///
/// \code
///     ./example2 [ -fixed-order [ -order O ] | -all-order [ -non-recursive ] [ -t T ] ] 
///                [ -q | -g ] [ -seed S ] [ -nev N ] [ -eps E ] [ -nbin B ]
///                [ -fcut F ] [ -nfilt M ] [ -output FN]
///               
/// \endcode
///
/// The code can be run to give fixed order results (-fixed-order),
/// specifically the coefficient of t^O, or all-order results
/// (-all-order) for a given value of t=T.
///
/// You can specify whether you start with a quark (default) or a
/// gluon with the -q and -g options.
///
/// Technical parameters:
///
/// - -non-recursive specifies an alternative all-order implementation
/// - -nev N         indicates that the program should generate N events (default 10^4)
/// - -eps E         specifies the value of the cutoff: E < z < 1-E.
/// - -seed S        specifies the random seed (default 1)
/// - -nbin B        indicates the number of bins for histograms
/// - -fcut F        specifies the treshold value F < z for trimming
/// - -nfilt M       specifies the number of hardest subjets for filtering
/// - -output FN     indicates the name of output file (default is to std::cout)
/// 

#include "CmdLine.hh"
#include "NLOHist.hh"
#include "AverageAndError.hh"
#include "Observable.hh"
#include "Output.hh"
#include "Generator.hh"
#include "AllOrderGeneratorFlat.hh"
#include "FixedOrderGenerator.hh"
#include "AllOrderGeneratorRecursive.hh"

using namespace std;

int main(int argc, char ** argv) {
  CmdLine cmdline(argc,argv);
  string filename = cmdline.value<string>("-output","");
  int seed = cmdline.value("-seed",1);
  Parton::Flavour flav = Parton::q;
  if (cmdline.present("-q")) flav = Parton::q;
  if (cmdline.present("-g")) flav = Parton::g;
  int order = cmdline.value("-order",2);
  double nev = cmdline.value("-nev",1e4);
  double eps = cmdline.value("-eps",1e-3);
  double t = cmdline.value("-t",0.5);
  int nbin = cmdline.value("-nbin",10);
  int nfilt = cmdline.value("-nfilt",2);
  double fcut = cmdline.value("-fcut",0.05);
  string command=cmdline.command_line();
  string time=cmdline.time_stamp();

  double CA = cmdline.value("-CA",3.0);
  double CF = cmdline.value("-CF",4.0/3.0);
  double TR = cmdline.value("-TR",0.5);
  double nf = cmdline.value("-nf",5.0);

  cout << "# " << command << "\n# " << time << endl;
  Output output(filename,command,time);

  Generator * generator;
  if (cmdline.present("-fixed-order") && !cmdline.present("-all-order")) {
    generator = new FixedOrderGenerator(seed,flav,order, CA, CF, TR, nf);
  }
  else if (cmdline.present("-non-recursive")) {
    generator = new AllOrderGeneratorFlat(seed,flav,eps,t, CA, CF, TR, nf);
  }
  else {
    generator = new AllOrderGeneratorRecursive(seed,flav,eps,t, CA, CF, TR, nf);
  }

  cmdline.assert_all_options_used();
  
  // Define averages and histograms
  map<string,AverageAndError> averages;
  vector<NLOHist> histos(2);
  histos[0].declare("x max",0.0,1.0+1e-9,nbin);
  histos[1].declare("x inclusive",0.0,1.0+1e-9,nbin);
  
  // quarkhardest(ev) returns 1 if hardest parton is a quark
  FlavHardest quarkhardest(Parton::q);
  // Filter the nfilt hardest subjets of an event
  Filter filter(nfilt);
  // Trim event with momentum treshold fcut
  Trim trim(fcut);
  

  for (int iev = 0; iev < nev; iev++) {
    // events may come in "groups" (like for a NLO program);
    // generate the next group
    generator->next_event_group();

    // get a reference to the next group of events
    const vector<Event> & events = generator->events();

    // calculate observables
    averages["delta z"]   += average_over_group(events,deltaxmax);
    averages["log z"]     += average_over_group(events,logxmax);
    averages["prob quark"]+= average_over_group(events,quarkhardest);
    averages["z^6"]       += average_over_group(events,xmax6);
    averages["trimmed z"] += average_over_group(events,trim);
    averages["filtered z"]+= average_over_group(events,filter);
    add_group_to_histogram(histos[0],events,xmax);
    add_group_to_histogram(histos[1],events,xinclusive);

    for(vector<NLOHist>::iterator it=histos.begin();
	it != histos.end(); ++it)
      it->collate_event();
    
    if(output.writing_step(iev)) {
      output.write(averages,histos,iev);
      cout << "# Currently at " << iev << " out of "
	   << nev << " events." << endl;
    }
    
    // Reset to initial configuration for next event
    generator->reset();
  }
  output.write(averages,histos,nev);
  cout << "# Simulation completed" << endl;
}

