#!/usr/bin/env python
#
# Python script to generate the uncertainty bands for (N)NLO and matched results

usage="""

  ./get-bands.py [-R R] [-y y] [-K Kfactor] [-Rm Rm] [-R0 R0]

"""


import cmdline
from hfile import *
import math
import numpy
import re
import sys

nloFile='atlas-1410.8857-nlo-CT10.res'
nlo3jFile='atlas-1410.8857-nlo-CT10-threejet.res'
pdf="CT10"


def main():

    print "# " + cmdline.cmdline()
    R = cmdline.value("-R", 0.4)
    R0 = cmdline.value("-R0", 1.0)
    y = cmdline.value("-y", 0.0)

    # K factor to study R-independent NNLO term
    Kfact = cmdline.value("-K",-1.0)

    # Rm to set matching radius where sigma^NNLO=sigma^NLO
    # (similar to K factor)
    Rm = cmdline.value("-Rm",1.0)
    
    # set up our list of scales
    scales = []
    scales.append((1.0, 1.0))
    scales.append((1.0, 2.0))
    scales.append((1.0, 0.5))
    scales.append((0.5, 1.0))
    scales.append((2.0, 1.0))
    scales.append((0.5, 0.5))
    scales.append((2.0, 2.0))
    
    central=scaleName(1.0, 1.0)
    loCol  = 4 - 1
    nloCol = 6 - 1

    # get column names and mapping back from name to index
    colnames = []
    colnames.append('ptlo'        )
    colnames.append('ptmd'        )
    colnames.append('pthi'        )
    colnames.append('lo_exact'    )
    colnames.append('nlo_exact'   )
    colnames.append('nlo_exact_R0')
    colnames.append('nlo_exact_diff')
    colnames.append('nnlo_exact_diff')
    colnames.append('nlo_mult'    )
    colnames.append('pseudo_nnlo')
    colnames.append('resummed'    )
    colnames.append('matched'   )
    colnames.append('matched_nnlo')
    colnames.append('lo_smallr'   )
    colnames.append('nlo_smallr'  )
    colnames.append('nnlo_smallr' )
    colnames.append('nnnlo_smallr')
    colnames.append('nlo_residue' )
    colnames.append('nnlo_residue')
    colnames.append('resummed_eff')
    colnames.append('fixedord_eff')
    colnames.append('nnlo_mult')
    colnames.append('fixedord_eff_nnlo')

    col = {}
    for index in range(0,len(colnames)):
        col[colnames[index]] = index
    
    # collect the nlo, resummed and matched results all into one place
    allres = {}
    allres_min = {}
    allres_max = {}
    for scale in scales:
        nlojet = get_array(nloFile, nloLabel(y, R, *scale))
        nlojetR0 = get_array(nloFile, nloLabel(y, R0, *scale))
        nlojetRm = get_array(nloFile, nloLabel(y, Rm, *scale))
        if (R!=1.0):
            nlo3jet = get_array(nlo3jFile, nloLabel(y, R, *scale))
        else:
            nlo3jet = get_array(nlo3jFile, nloLabel(y, 0.1, *scale))
            nlo3jet[:,3:6] = 0.0 * nlo3jet[:,3:6]
        nlo3jetRm = numpy.empty(nlo3jet.shape)
        nlo3jetR0 = numpy.empty(nlo3jet.shape)
        if (R0 != 1.0):
            nlo3jetR0 = get_array(nlo3jFile, nloLabel(y, R0, *scale))
        if ((Rm > 0.0) and (Rm != 1.0)):
            nlo3jetRm = get_array(nlo3jFile, nloLabel(y, Rm, *scale))
        resum  = get_array(resumFile(y,R,R0,*scale), "microjet.spectrum")
        expnd  = get_array(resumFile(y,R,R0,*scale), "flavour-summed.+expanded.in.alphas")

        # get the dimensions of our main results array
        npt  = len(resum[:,0])
        ncol = len(col)
        scName = scaleName(*scale)
        allres[scName] = numpy.empty([npt, ncol])

        # point to it for convenience
        pointer = allres[scName]

        # fill up the columns we take from elsewhere
        pointer[:,col['ptlo']] = nlojet[:,0]
        pointer[:,col['ptmd']] = nlojet[:,1]
        pointer[:,col['pthi']] = nlojet[:,2]

        pointer[:,col['lo_exact' ]] = nlojet[:,3]
        pointer[:,col['nlo_exact']] = nlojet[:,5]
        pointer[:,col['nlo_exact_R0']] = nlojetR0[:,5]
        pointer[:,col['nlo_exact_diff' ]] = nlo3jet[:,3]
        pointer[:,col['nnlo_exact_diff']] = nlo3jet[:,5]
        
        # if Rm>0.0 we want to have e2R be set to 0 at R=Rm instead of default R=1
        # so subtract the (R=Rm - R=1.0) piece which is contained in nlo3jetR0
        if ((Rm>0.0) and (Rm!=1.0)):
            pointer[:,col['nlo_exact_diff' ]] -= nlo3jetRm[:,3]
            pointer[:,col['nnlo_exact_diff']] -= nlo3jetRm[:,5]

        pointer[:,col['resummed'] ]  = resum[:,4]
        pointer[:,col['lo_smallr']:(col['nnnlo_smallr']+1)] = expnd[:,2:]

        # finally do the matching
        lo_exact    = pointer[:,col['lo_exact'  ]]
        nlo_exact   = pointer[:,col['nlo_exact' ]]  # the LO plus NLO piece
        nlo_R0      = pointer[:,col['nlo_exact_R0']]  # the LO plus NLO piece
        nlo_Rm      = nlojetR0[:,5]
        nlo_exact_diff = pointer[:,col['nlo_exact_diff']] # the full NLO, diff between R0=1.0 and R
        nnlo_exact_diff = pointer[:,col['nnlo_exact_diff']] # the full NNLO, diff between R0=1.0 and R
        lo_smallr   = pointer[:,col['lo_smallr' ]]
        nlo_smallr  = pointer[:,col['nlo_smallr']]  # the pure NLO piece
        nnlo_smallr = pointer[:,col['nnlo_smallr']] # the pure NNLO piece
        resummed    = pointer[:,col['resummed'  ]]
        nlo_exact_diff_R0  = nlo3jetR0[:,3] # difference between R=1 and R0 at NLO
        nnlo_exact_diff_R0 = nlo3jetR0[:,5] # difference between R=1 and R0 at NNLO
        
        # a nlo multiplicative scheme
        pointer[:,col['nlo_mult']] = (1 + (nlo_exact - nlo_R0)/lo_exact) * (nlo_R0)

        # now define some shorthands
        e0   = lo_exact
        e1R  = nlo_exact - lo_exact
        e1R0 = nlo_R0 - lo_exact
        e1Rm = nlo_Rm - lo_exact
        e2R  = nnlo_exact_diff - nlo_exact_diff
        e2R0 = 0.0 # this will become non-zero when using a K-factor or R0 = 1.5

        if (R0 != 1.0):
            # if R0 = 1.5, then e2R0 is not zero
            e2R0 = nnlo_exact_diff_R0 - nlo_exact_diff_R0
            
        # Explore impact of NNLO K-factor
        # pseudo nnlo: + nloR0*(K-1)
        # nnlo mult  : + (1+(e1R-e1R0)/e0+...) * nloR0*(K-1)
        # nnlo mA : + R/e0 * nloR0*(K-1)
        # nnlo mB : + R/e0 * (1+(e1R-e1R0-r1R)/e0+...) * nloR0*(K-1)
        if (Kfact>0.0): # Kfact is set to -1 by default
            kval = (Kfact-1.0)*(lo_exact + e1Rm) # rescale by NLO(R_m)
            e2R0 += kval
            e2R  += kval
        r1R  = nlo_smallr
        r2R  = nnlo_smallr

        # now do the actual matching to NLO
        pointer[:,col['matched']] = nlo_R0 * resummed/lo_smallr * (1 + (e1R - e1R0 - r1R)/e0)
        
        # NNLO multiplicative scheme
        pointer[:,col['matched_nnlo']] = (resummed/lo_smallr) * (nlo_R0+e2R0) * \
                (1.0 + (e1R - e1R0 - r1R)/e0 + (e2R - e2R0 - r2R)/e0 \
                 - (r1R/e0) * (e1R - r1R)/e0 - (e1R0/e0) * (e1R - e1R0 -r1R)/e0)

        # a "Pseudo" NNLO, assuming that NNLO(R=1) = NLO(R=1)
        pointer[:,col['pseudo_nnlo']] = e0 + e1R + e2R
        
        # a nnlo multiplicative scheme
        pointer[:,col['nnlo_mult']] = (1/e0) * (e0 + e1R + e2R - e1R0 - e2R0 - e1R0*(e1R-e1R0)/e0) * (nlo_R0+e2R0)
        pointer[:,col['fixedord_eff_nnlo']] = (e0 + e1R + e2R - e1R0 - e2R0 - e1R0*(e1R-e1R0)/e0)/e0
        
        # and the elements needed for the efficiency method (matched only)
        pointer[:,col['resummed_eff']] = resummed/lo_smallr * (1 + (e1R - e1R0 - r1R)/e0)
        pointer[:,col['fixedord_eff']] = 1 + (nlo_exact - nlo_R0) / lo_exact
        pointer[:,col['nlo_residue']]  = (nlo_exact-nlo_smallr) 
        pointer[:,col['nnlo_residue']] = (resummed/lo_smallr) * \
                                    (1.0 + (e1R - e1R0 - r1R)/e0 + (e2R - e2R0 - r2R)/e0 \
                                     - (r1R/e0) * (e1R - r1R)/e0 - (e1R0/e0) * (e1R - e1R0 -r1R)/e0)
        
    allres_min = minimum(allres)
    allres_max = maximum(allres)
    if (R<0.1):
        print "# y{y:3.1f}-R{R:3.2f}-R0{R0:3.1f}-allscales".format(y=y,R=R,R0=R0)
    else:
        print "# y{y:3.1f}-R{R:3.1f}-R0{R0:3.1f}-allscales".format(y=y,R=R,R0=R0)
    print_minmax(colnames, allres[scaleName(1.0,1.0)], allres_min, allres_max)
    print

#----------------------------------------------------------------------
def print_minmax(colnames, centralarray, minarray, maxarray):
    newnames = []
    newarray = []
    for i in range(0,len(colnames)):
        colname = colnames[i]
        if (re.match(r'^pt',colname)):
            # if the column name starts with pt, we don't need min and max values
            newnames.append(colname)
            newarray.append(centralarray[:,i])
        elif not ((re.search('small',colname)) or (re.search('diff',colname))):
            # we don't need columns with small-R expansion and fixed-order differences anymore,
            # so ignore these columns.

            # otherwise we get three columns per original column
            newnames.append(colname+"_central")
            newnames.append(colname+"_min")
            newnames.append(colname+"_max")
            # and now the arrays
            newarray.append(centralarray[:,i])
            newarray.append(minarray[:,i])
            newarray.append(maxarray[:,i])

    # build up column names (again...)
    col = {}
    for index in range(0,len(colnames)):
        col[colnames[index]] = index

    # and generate the alternative uncertainties for matched NLO
    # (like efficiency method: add in quadrature efficiency and total xsct uncertainties)
    newnames.append('matched_alt_min')
    newnames.append('matched_alt_max')
    norm_nlo = centralarray[:,col['matched']]
    matched_alt_min = (1-numpy.sqrt( (minarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2
                        + (minarray[:,col['resummed_eff']]/centralarray[:,col['resummed_eff']] - 1)**2 ))*norm_nlo
    matched_alt_max = (1+numpy.sqrt( (maxarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2
                        + (maxarray[:,col['resummed_eff']]/centralarray[:,col['resummed_eff']] - 1)**2 ))*norm_nlo
    newarray.append(matched_alt_min)
    newarray.append(matched_alt_max)
    
    # now generate alternative uncertainties for nnlo matching
    newnames.append('matched_nnlo_alt_min')
    newnames.append('matched_nnlo_alt_max')
    norm_nnlo = centralarray[:,col['matched_nnlo']]
    matched_nnlo_alt_min = (1-numpy.sqrt( (minarray[:,col['nnlo_residue']]/centralarray[:,col['nnlo_residue']] - 1)**2
                                    + (minarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2 ))*norm_nnlo
    matched_nnlo_alt_max = (1+numpy.sqrt( (maxarray[:,col['nnlo_residue']]/centralarray[:,col['nnlo_residue']] - 1)**2
                                    + (maxarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2 ))*norm_nnlo
    newarray.append(matched_nnlo_alt_min)
    newarray.append(matched_nnlo_alt_max)

    # now generate alternative uncertainties for NLO mult. scheme
    newnames.append('nlo_mult_alt_min')
    newnames.append('nlo_mult_alt_max')
    norm_nlom = centralarray[:,col['nlo_mult']]
    nlo_mult_alt_min = (1-numpy.sqrt( (minarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2
                                    + (minarray[:,col['fixedord_eff']]/centralarray[:,col['fixedord_eff']] - 1)**2 ))*norm_nlom
    nlo_mult_alt_max = (1+numpy.sqrt( (maxarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2
                                    + (maxarray[:,col['fixedord_eff']]/centralarray[:,col['fixedord_eff']] - 1)**2 ))*norm_nlom
    newarray.append(nlo_mult_alt_min)
    newarray.append(nlo_mult_alt_max)

    # now generate alternative uncertainties for NNLO mult. scheme
    newnames.append('nnlo_mult_alt_min')
    newnames.append('nnlo_mult_alt_max')
    norm_nnlom = centralarray[:,col['nnlo_mult']]
    nnlo_mult_alt_min = (1-numpy.sqrt( (minarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2
                        + (minarray[:,col['fixedord_eff_nnlo']]/centralarray[:,col['fixedord_eff_nnlo']] - 1)**2 ))*norm_nnlom
    nnlo_mult_alt_max = (1+numpy.sqrt( (maxarray[:,col['nlo_exact_R0']]/centralarray[:,col['nlo_exact_R0']] - 1)**2
                        + (maxarray[:,col['fixedord_eff_nnlo']]/centralarray[:,col['fixedord_eff_nnlo']] - 1)**2 ))*norm_nnlom
    newarray.append(nnlo_mult_alt_min)
    newarray.append(nnlo_mult_alt_max)

    # build up column names (again...)
    colnew = {}
    for index in range(0,len(newnames)):
        colnew[newnames[index]] = index
    # remove all the columns which are not needed in the output
    # and are only there for intermediate steps
    rmlist = ['nlo_exact_R0','resummed','nlo_residue','nnlo_residue','resummed_eff','fixedord_eff','fixedord_eff_nnlo']
    for rmel in rmlist:
        # remove three times to get rid of central, min and max values
        del newnames[colnew[rmel+'_central']]
        del newnames[colnew[rmel+'_central']]
        del newnames[colnew[rmel+'_central']]
        del newarray[colnew[rmel+'_central']]
        del newarray[colnew[rmel+'_central']]
        del newarray[colnew[rmel+'_central']]
        # build up column names (again...)
        colnew = {}
        for index in range(0,len(newnames)):
            colnew[newnames[index]] = index

        
    print " ".join(newnames)
    print reformat(*newarray)
    
#----------------------------------------------------------------------
def resumFile(y, R, R0, xmur, xmuf):
    "return the name of the resummed results file"
    if (R < 0.1):
        return ("resum/y{y0:3.1f}-{y1:3.1f}-R{R:4.2f}-R0{R0:3.1f}-xmur{xmur:3.1f}-xmuf{xmuf:3.1f}-{pdf}.rebin".format
                (y0=y, y1=y+0.5, R=R, R0=R0, xmur=xmur, xmuf=xmuf, pdf=pdf))
    return ("resum/y{y0:3.1f}-{y1:3.1f}-R{R:3.1f}-R0{R0:3.1f}-xmur{xmur:3.1f}-xmuf{xmuf:3.1f}-{pdf}.rebin".format
            (y0=y, y1=y+0.5, R=R, R0=R0, xmur=xmur, xmuf=xmuf, pdf=pdf))
    
#----------------------------------------------------------------------
def scaleName(xmuR, xmuF):
    return "xmuR{0:3.1f}-xmuF{1:3.1f}".format(xmuR, xmuF)

#----------------------------------------------------------------------
def nloLabel(y, R, xmuR, xmuF):
    "returns the label to look for in the NLO file, as a function of rapidity, R, xmuR and xmuF"
    return ("dsigma_dpt-y_"+gregFormat(y)+"_"+gregFormat(y+0.5)
            +"-R_"+gregFormat(R)
            +"-mur_"+gregFormat(xmuR)+"-muf_"+gregFormat(xmuF))


#----------------------------------------------------------------------
def gregFormat(x):
    "Return integers formatted without a decimal point, other numbers with, down to 0.01"
    if (int(x) == x)      : return "{0:1.0f}".format(x)
    if (int(10*x) == 10*x): return "{0:3.1f}".format(x)
    else                  : return "{0:4.2f}".format(x)

if __name__ == '__main__':
    main()
